#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
nmrstarlib.nmrstarlib
~~~~~~~~~~~~~~~~~~~~~

This module provides the :class:`~nmrstarlib.nmrstarlib.StarFile` class
that stores the data from a single NMR-STAR file in the form of an
:py:class:`~collections.OrderedDict`. Data can be accessed directly from the
:class:`~nmrstarlib.nmrstarlib.StarFile` instance using bracket
accessors.

The NMR-STAR format is a hierachical dictionary containing data on
NMR experiments. The data is divided into a series of "saveframes"
which each contain a number of key-value pairs and "loops".

Each saveframe has a unique name, which is used as the key in
the dictionary, corresponding to another dictionary containing the
information in the saveframe.  Since loops in NMR-Star format do
not have names, the keys for them inside the saveframe dictionary
are simply loop_0, loop_1, etc.
"""

import sys
import os
import io
import zipfile
import bz2
import gzip
import tarfile
import json
import urllib.request
import urllib.error
from collections import OrderedDict

from . import bmrblex


BMRB_REST = "http://rest.bmrb.wisc.edu/bmrb/NMR-STAR3/"
VERBOSE = False
NMRSTAR_VERSION = "3"
NMRSTAR_CONSTANTS = {}


class StarFile(OrderedDict):
    """StarFile class that stores the data from a single NMR-STAR file in the form of an
    :py:class:`~collections.OrderedDict`."""

    def __init__(self, source="", frame_categories=None):
        """`StarFile` initializer. Leave `frame_categories` as `None` to
        read everything.  Otherwise it can be a list of saveframe
        categories to read, skipping the rest.

        :param str source: Source `StarFile` created from: local file or URL address.
        :param list frame_categories: List of saveframe names.
        """
        super().__init__(self)
        self.source = source
        self._frame_categories = frame_categories
        self.bmrbid = ""

    def read(self, filehandle):
        """Read data into a :class:`~nmrstarlib.nmrstarlib.StarFile` instance.

        :param filehandle: file-like object.
        :type filehandle: :py:class:`io.TextIOWrapper`, :py:class:`gzip.GzipFile`,
                          :py:class:`bz2.BZ2File`, :py:class:`zipfile.ZipFile`
        :return: None
        :rtype: None
        """
        inputstr = filehandle.read()
        nmrstar_str = self._is_nmrstar(inputstr)
        json_str = self._is_json(inputstr)

        if inputstr == "" or inputstr == b"":
            pass
        elif nmrstar_str:
            lexer = bmrblex.bmrblex(nmrstar_str)
            self._build_starfile(lexer)
        elif json_str:
            self.update(json_str)
            self.bmrbid = self["data"]
        else:
            raise TypeError("Unknown file format")
        filehandle.close()

    def write(self, filehandle, fileformat):
        """Write :class:`~nmrstarlib.nmrstarlib.StarFile` data into file.

        :param filehandle: file-like object.
        :type filehandle: :py:class:`io.TextIOWrapper`
        :param str fileformat: Format to use to write data: `nmrstar` or `json`.
        :return: None
        :rtype: None
        """
        try:
            if fileformat == "json":
                json_str = self._to_json()
                filehandle.write(json_str)
            elif fileformat == "nmrstar":
                nmrstar_str = self._to_nmrstar()
                filehandle.write(nmrstar_str)
            else:
                raise TypeError("Unknown file format.")
        except IOError:
            raise IOError('"filehandle" parameter must be writable.')
        filehandle.close()

    def writestr(self, fileformat):
        """Write :class:`~nmrstarlib.nmrstarlib.StarFile` data into string.

        :param str fileformat: Format to use to write data: `nmrstar` or `json`.
        :return: String representing the :class:`~nmrstarlib.nmrstarlib.StarFile` instance.
        :rtype: str
        """
        try:
            if fileformat == "json":
                json_str = self._to_json()
                return json_str
            elif fileformat == "nmrstar":
                nmrstar_str = self._to_nmrstar()
                return nmrstar_str
            else:
                raise TypeError("Unknown file format.")
        except IOError:
            raise IOError('"filehandle" parameter must be writable.')

    def _build_starfile(self, lexer):
        """Build :class:`~nmrstarlib.nmrstarlib.StarFile` object.

        :param lexer: instance of the BMRB lexical analyzer class.
        :type lexer: :class:`~nmrstarlib.bmrblex.bmrblex`
        :return: instance of :class:`~nmrstarlib.nmrstarlib.StarFile`.
        :rtype: :class:`~nmrstarlib.nmrstarlib.StarFile`
        """
        odict = self
        token = lexer.get_token()

        while token != "":
            try:
                if token[0:5] == "save_":
                    # name = token[5:]
                    name = token
                    frame = self._build_saveframe(lexer)
                    if frame:
                        odict[name] = frame
                    
                elif token[0:5] == "data_":
                    self.bmrbid = token[5:]
                    self["data"] = self.bmrbid
                else:
                    # print('{} Error: Invalid token {}'.format(lexer.error_leader(), token), file=sys.stderr)
                    print("Error: Invalid token {}".format(token), file=sys.stderr)
                    print("In _build_starfile try block", file=sys.stderr)  # DEBUG
                    # raise InvalidToken("{} {}".format(lexer.error_leader(), token))
                    raise InvalidToken("{}".format(token))

            except IndexError:
                # print('{} Error: Invalid token {}'.format(lexer.error_leader(), token), file=sys.stderr)
                print("Error: Invalid token {}".format(token), file=sys.stderr)
                print("In _build_starfile except block", file=sys.stderr)  # DEBUG
                raise
            finally:
                token = lexer.get_token()
        return self

    def _build_saveframe(self, lexer):
        """Build NMR-STAR file saveframe.

        :param lexer: instance of the BMRB lexical analyzer class.
        :type lexer: :class:`~nmrstarlib.bmrblex.bmrblex`
        :return: Saveframe dictionary.
        :rtype: :py:class:`collections.OrderedDict`
        """
        odict = OrderedDict()
        loopcount = 0

        token = lexer.get_token()
        while token != "save_":
            try:
                if token[0] == "_":
                    # This strips off the leading underscore of tagnames for readability 
                    odict[token[1:]] = lexer.get_token()

                    # Skip the saveframe if it's not in the list of wanted categories
                    if self._frame_categories:
                        if token == "_Saveframe_category" and odict[token[1:]] not in self._frame_categories:
                            raise SkipSaveFrame()

                elif token == "loop_":
                    odict["loop_{}".format(loopcount)] = self._build_loop(lexer)
                    loopcount += 1

                else:
                    # print('{} Error: Invalid token {}'.format(lexer.error_leader(), token), file=sys.stderr)
                    print("Error: Invalid token {}".format(token), file=sys.stderr)
                    print("In _build_saveframe try block", file=sys.stderr)  # DEBUG
                    # raise InvalidToken("{} {}".format(lexer.error_leader(), token))
                    raise InvalidToken("{}".format(token))

            except IndexError:
                # print("{} Error: Invalid token {}".format(lexer.error_leader(), token), file=sys.stderr)
                print("Error: Invalid token {}".format(token), file=sys.stderr)
                print("In _build_saveframe except block", file=sys.stderr)  # DEBUG
                raise
            except SkipSaveFrame:
                self._skip_saveframe(lexer)
                odict = None
            finally:
                if odict is None:
                    token = "save_"
                else:
                    token = lexer.get_token()
        return odict

    def _skip_saveframe(self, lexer):
        """Skip entire saveframe - keep emitting tokens until the end of saveframe.

        :param lexer: instance of the BMRB lexical analyzer class.
        :type lexer: :class:`~nmrstarlib.bmrblex.bmrblex`
        :return: None
        :rtype: None
        """
        token = ''
        while token != 'save_':
            token = lexer.get_token()

    def _build_loop(self, lexer):
        """Build saveframe loop.

        :param lexer: instance of BMRB lexical analyzer class.
        :type lexer: :class:`~nmrstarlib.bmrblex.bmrblex`
        :return: Fields and values of the loop.
        :rtype: tuple
        """
        fields = []
        values = []

        token = lexer.get_token()
        while token[0] == "_":
            fields.append(token[1:])
            token = lexer.get_token()

        while token != "stop_":
            values.append(token)
            token = lexer.get_token()

        assert (len(values)/len(fields)).is_integer(), \
            "Error in loop construction: number of fields must be equal to number of values."

        # divide list of loop values into chunks corresponding to fields
        # values = [values[i:i+len(fields)] for i in range(0, len(values), len(fields))]
        # return fields, values

        values = [OrderedDict(zip(fields, values[i:i + len(fields)])) for i in range(0, len(values), len(fields))]
        return fields, values

    def print_starfile(self, f=sys.stdout, format="nmrstar", tw=3):
        """Print :class:`~nmrstarlib.nmrstarlib.StarFile` into a file or stdout.

        :param io.StringIO f: writable file-like stream.
        :param str format: Format to use: `nmrstar` or `json`.
        :param int tw: Tab width.
        :return: None
        :rtype: None
        """
        if format is "nmrstar":
            for saveframe in self.keys():
                if saveframe == "data":
                    print("{}_{}\n".format(saveframe, self[saveframe]), file=f)
                else:
                    print("{}".format(saveframe), file=f)
                    self.print_saveframe(saveframe, f, format, tw)
                    print("save_\n", file=f)

        elif format is "json":
            print(self._to_json(), file=f)

    def print_saveframe(self, sf, f=sys.stdout, format="nmrstar", tw=3):
        """Print saveframe into a file or stdout.
        We need to keep track of how far over everything is tabbed. The "tab width"
        variable tw does this for us.

        :param str sf: Saveframe name.
        :param io.StringIO f: writable file-like stream.
        :param str format: Format to use: `nmrstar` or `json`.
        :param int tw: Tab width.
        :return: None
        :rtype: None
        """
        if format is "nmrstar":
            if sf == "data":
                print("{}_{}\n".format(sf, self[sf]), file=f)
            else:
                for sftag in self[sf].keys():
                    # handle the NMR-Star "long string"
                    if self[sf][sftag][0] == ";":
                        print(tw*" ", "_{}".format(sftag), file=f)
                        print("{}".format(self[sf][sftag]), file=f)

                    # handle loops
                    elif sftag[:5] == "loop_":
                        print(tw*" ", "loop_", file=f)
                        self.print_loop(sf, sftag, f, format, tw * 2)
                        print(tw*" ", "stop_", file=f)
                    else:
                        print(tw*" ", "_{}\t {}".format(sftag, self[sf][sftag]), file=f)

        elif format is "json":
            print(json.dumps(self[sf], sort_keys=False, indent=4), file=f)

    def print_loop(self, sf, sftag, f=sys.stdout, format="nmrstar", tw=3):
        """Print loop into a file or stdout.

        :param str sf: Saveframe name.
        :param str sftag: Saveframe tag, i.e. field name.
        :param io.StringIO f: writable file-like stream.
        :param str format: Format to use: `nmrstar` or `json`.
        :param int tw: Tab width.
        :return: None
        :rtype: None
        """
        if format is "nmrstar":
            # First print the fields
            for field in self[sf][sftag][0]:
                print(tw*" ", "_{}".format(field), file=f)

            # Then print the values
            for valuesdict in self[sf][sftag][1]:
                line = tw*" " + " ".join(valuesdict.values())
                print(line, file=f)

        elif format is "json":
            print(json.dumps(self[sf][sftag], sort_keys=False, indent=4), file=f)

    def _to_json(self):
        """Save :class:`~nmrstarlib.nmrstarlib.StarFile` into JSON string.

        :return: JSON string.
        :rtype: str
        """
        return json.dumps(self, sort_keys=False, indent=4)

    def _to_nmrstar(self):
        """Save :class:`~nmrstarlib.nmrstarlib.StarFile` NMR-STAR format string.

        :return: NMR-STAR string.
        :rtype: str
        """
        nmrstar_str = io.StringIO()
        self.print_starfile(nmrstar_str)
        return nmrstar_str.getvalue()

    @staticmethod
    def _is_nmrstar(string):
        """Test if input string is in NMR-STAR format.

        :param string: Input string.
        :type string: str or bytes
        :return: Input string if in NMR-STAR format or False otherwise.
        :rtype: str or False
        """
        if string[0:5] == "data_" or string[0:5] == b"data_":
            return string
        return False

    @staticmethod
    def _is_json(string):
        """Test if input string is in JSON format.

        :param string: Input string.
        :type string: str or bytes
        :return: Input string if in JSON format or False otherwise.
        :rtype: str or False
        """
        try:
            if isinstance(string, bytes):
                json_str = json.loads(string.decode("utf-8"), object_pairs_hook=OrderedDict)
            elif isinstance(string, str):
                json_str = json.loads(string, object_pairs_hook=OrderedDict)
            else:
                raise TypeError("Expecting <class 'str'> or <class 'bytes'>, but {} was passed".format(type(string)))
            return json_str
        except ValueError:
            return False

    def chem_shifts_by_residue(self, aminoacids=None, atoms=None, nmrstarversion="3"):
        """Organize chemical shifts by amino acid residue.

        :param list aminoacids: List of aminoacids three-letter codes.
        :param list atoms: List of BMRB atom type codes.
        :return: List of OrderedDict per each chain
        :rtype: list of OrderedDict
        """
        this_directory = os.path.dirname(__file__)
        if nmrstarversion == "2":
            config_filepath = os.path.join(this_directory, 'conf/constants_nmrstar2.json')
        elif nmrstarversion == "3":
            config_filepath = os.path.join(this_directory, 'conf/constants_nmrstar3.json')
        else:
            config_filepath = os.path.join(this_directory, 'conf/constants_nmrstar3.json')
        with open(config_filepath, "r") as infile:
            update_constants(infile)

        chemshifts_loop = NMRSTAR_CONSTANTS["chemshifts_loop"]
        aminoacid_seq_id = NMRSTAR_CONSTANTS["aminoacid_seq_id"]
        aminoacid_code = NMRSTAR_CONSTANTS["aminoacid_code"]
        atom_code = NMRSTAR_CONSTANTS["atom_code"]
        chemshift_value = NMRSTAR_CONSTANTS["chemshift_value"]

        chains = []
        for saveframe in self:
            if saveframe == "data":
                continue
            else:
                for ind in self[saveframe].keys():
                    if ind.startswith("loop_"):
                        if list(self[saveframe][ind][0]) == chemshifts_loop:
                            chemshifts_dict = OrderedDict()
                            for entry in self[saveframe][ind][1]:
                                residueid = (entry[aminoacid_seq_id], entry[aminoacid_code])
                                chemshifts_dict.setdefault(residueid, OrderedDict())
                                chemshifts_dict[residueid][entry[atom_code]] = entry[chemshift_value]
                            chains.append(chemshifts_dict)

        if aminoacids:
            for chemshifts_dict in chains:
                for aa in list(chemshifts_dict.keys()):
                    if aa[1].upper() not in aminoacids:
                        chemshifts_dict.pop(aa)

        if atoms:
            for chemshifts_dict in chains:
                for resonances_dict in chemshifts_dict.values():
                    for at in list(resonances_dict.keys()):
                        if at.upper() not in atoms:
                            resonances_dict.pop(at)
        return chains


def update_constants(filehandle):
    """Update constants related to NMR-STAR format, e.g. field names.

    :param str filehandle: JSON file that contains information about NMR-STAR format.
    :return: None
    :rtype: None
    """
    newconstants = json.loads(filehandle.read())
    NMRSTAR_CONSTANTS.update(newconstants)


def _generate_filenames(sources):
    """Generate filenames.

    :param list sources: List of strings representing path to file(s).
    :return: Path to file(s).
    :rtype: str
    """
    for source in sources:
        if os.path.isdir(source):
            for path, dirlist, filelist in os.walk(source):
                for fname in filelist:
                    if GenericFilePath.is_compressed(fname):
                        if VERBOSE:
                            print("Skipping compressed file: {}".format(os.path.abspath(fname)))
                        continue
                    else:
                        yield os.path.join(path, fname)
        elif os.path.isfile(source):
            yield source
        elif source.isdigit():
            yield BMRB_REST + source
        elif GenericFilePath.is_url(source):
            yield source
        else:
            raise TypeError("Unknown file source.")


def _generate_handles(filenames):
    """Open a sequence of filenames one at time producing file objects.
    The file is closed immediately when proceeding to the next iteration.

    :param generator filenames: Generator object that yields the path to each file, one at a time.
    :return: Filehandle to be processed into a :class:`~nmrstarlib.nmrstarlib.StarFile` instance.
    """
    for fname in filenames:
        if VERBOSE:
            print("Processing file: {}".format(os.path.abspath(fname)))
        path = GenericFilePath(fname)
        for filehandle, source in path.open():
            yield filehandle, source
            filehandle.close()


def read_files(sources):
    """Construct a generator that yields :class:`~nmrstarlib.nmrstarlib.StarFile` instances.

    :param list sources: List of strings representing path to file(s).
    :return: :class:`~nmrstarlib.nmrstarlib.StarFile` instance(s).
    :rtype: :class:`~nmrstarlib.nmrstarlib.StarFile`
    """
    filenames = _generate_filenames(sources)
    filehandles = _generate_handles(filenames)
    for fh, source in filehandles:
        starfile = StarFile(source)
        starfile.read(fh)
        yield starfile


class GenericFilePath(object):
    """`GenericFilePath` class knows how to open local files or files over URL."""

    def __init__(self, path):
        """Initialize path.

        :param str path: String representing a path to local file(s) or valid URL address of file(s).
        """
        self.path = path

    def open(self):
        """Generator that opens and yields filehandles using appropriate facilities:
        test if path represents a local file or file over URL, if file is compressed
        or not.

        :return: Filehandle to be processed into a :class:`~nmrstarlib.nmrstarlib.StarFile` instance.
        """
        is_url = self.is_url(self.path)
        compressiontype = self.is_compressed(self.path)

        if not compressiontype:
            if is_url:
                filehandle = urllib.request.urlopen(self.path)
            else:
                filehandle = open(self.path, "r")
            source = self.path
            yield filehandle, source
            filehandle.close()

        elif compressiontype:
            if is_url:
                response = urllib.request.urlopen(self.path)
                path = response.read()
                response.close()
            else:
                path = self.path

            if compressiontype == "zip":
                ziparchive = zipfile.ZipFile(io.BytesIO(path), "r") if is_url else zipfile.ZipFile(path)
                for name in ziparchive.infolist():
                    if not name.filename.endswith("/"):
                        filehandle = ziparchive.open(name)
                        source = self.path + "/" + name.filename
                        yield filehandle, source
                        filehandle.close()

            elif compressiontype in ("tar", "tar.bz2", "tar.gz"):
                tararchive = tarfile.open(fileobj=io.BytesIO(path)) if is_url else tarfile.open(path)
                for name in tararchive:
                    if name.isfile():
                        filehandle = tararchive.extractfile(name)
                        source = self.path + "/" + name.name
                        yield filehandle, source
                        filehandle.close()

            elif compressiontype == "bz2":
                filehandle = bz2.open(io.BytesIO(path)) if is_url else bz2.open(path)
                source = self.path
                yield filehandle, source
                filehandle.close()

            elif compressiontype == "gz":
                filehandle = gzip.open(io.BytesIO(path)) if is_url else gzip.open(path)
                source = self.path
                yield filehandle, source
                filehandle.close()

    @staticmethod
    def is_compressed(path):
        """Test if path represents compressed file(s).

        :param str path: Path to file(s).
        :return: String specifying compression type if compressed, "" otherwise.
        :rtype: str
        """
        if path.endswith(".zip"):
            return "zip"
        elif path.endswith(".tar.gz"):
            return "tar.gz"
        elif path.endswith(".tar.bz2"):
            return "tar.bz2"
        elif path.endswith(".gz"):
            return "gz"
        elif path.endswith(".bz2"):
            return "bz2"
        elif path.endswith(".tar"):
            return "tar"
        return ""

    @staticmethod
    def is_url(path):
        """Test if path represents a valid URL.

        :param str path: Path to file.
        :return: Valid :py:class:`~urllib.request.Request` object or False otherwise.
        :rtype: :py:class:`~urllib.request.Request` object or False
        """
        try:
            return urllib.request.Request(path)
        except ValueError:
            return False


class InvalidToken(Exception):
    def __init__(self, value):
        self.parameter = value

    def __str__(self):
        return repr(self.parameter)


class SkipSaveFrame(Exception):
    def __init__(self):
        pass

    def __str__(self):
        return 'Skipping save frame'


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
from pycallgraph import PyCallGraph
from pycallgraph.output import GraphvizOutput
from . import nmrstarlib


if __name__ == "__main__":
    # python3 -m nmrstarlib.profile_pycallgraph 18569

    source = sys.argv[1:]
    starfile_generator = nmrstarlib.read_files(source)

    with PyCallGraph(output=GraphvizOutput()):
        next(starfile_generator)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Routines for working with BMRB NMR-STAR format files.

This package includes the following modules:

``nmrstarlib``
    This module provides the :class:`~nmrstarlib.nmrstarlib.StarFile` class which is a python
    dictionary representation of a BMRB NMR-STAR file. Data can be accessed
    directly from the :class:`~nmrstarlib.nmrstarlib.StarFile` instance using bracket accessors.
    The :mod:`~nmrstarlib.nmrstarlib` module relies on the :mod:`~nmrstarlib.bmrblex` module
    for processing of tokens.

``bmrblex``
    This module provides the :class:`~nmrstarlib.bmrblex.bmrblex` class that is responsible for
    the syntax analysis of BMRB NMR-STAR files, processing word, number, single quoted,
    double quoted, multiline quoted BMRB tokens.

``converter``
    This module provides the :class:`~nmrstarlib.converter.Converter` class that is
    responsible for the conversion between NMR-STAR formated files and an equivalent
    JSONized file format.

``csviewer``
    This module provides the :class:`~nmrstarlib.csviewer.CSViewer` class that visualizes
    chemical shift values using the Graphviz (http://www.graphviz.org/) DOT Languge description.
"""

import os
from . import nmrstarlib

this_directory = os.path.dirname(__file__)

if nmrstarlib.NMRSTAR_VERSION == "3":
    config_filepath = os.path.join(this_directory, 'conf/constants_nmrstar3.json')
elif nmrstarlib.NMRSTAR_VERSION == "2":
    config_filepath = os.path.join(this_directory, 'conf/constants_nmrstar2.json')
else:
    config_filepath = os.path.join(this_directory, 'conf/constants_nmrstar3.json')

with open(config_filepath, "r") as infile:
    nmrstarlib.update_constants(infile)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
nmrstarlib.bmrblex
~~~~~~~~~~~~~~~~~~

A lexical analyzer class for BMRB STAR syntax.

This module is based on python ``shlex`` module modified to address specifics of BMRB NMR-STAR format.
The ``shlex`` class makes it easy to write lexical analyzers for simple syntaxes resembling
that of the Unix shell. Documentation: https://docs.python.org/3/library/shlex.html


Parsing rules:
--------------
   * Each word or number separated by whitespace characters is a separate BMRB token.
   * Each single quoted (') string is a separate BMRB token, it should start with single quote (')
     and end with single quote *always* followed by whitespace characters.
   * Each double quoted (") string is a separate BMRB token, it should start with double quote (")
     and end with double quote *always* followed by whitespace characters.
   * Single quoted and double quoted strings have to be processed separately.
   * Single quoted and double quoted strings are processed one character at a time.
   * Multiline strings starts with semicolon *always* followed by new line character and
     end with semicolon *always* followed by new line character.
   * Multiline strings are processed one line at a time.
"""

import sys
from collections import deque
from io import StringIO

__all__ = ["bmrblex"]

class bmrblex:
    """A lexical analyzer class for BMRB STAR syntax."""

    def __init__(self, instream=None):  # take infile out
        if isinstance(instream, str):
            instream = StringIO(instream)
        elif isinstance(instream, bytes):
            instream = StringIO(instream.decode("utf-8"))

        if instream is not None:
            self.instream = instream
        else:
            self.instream = sys.stdin

        self.eof = ''
        self.commenters = '#'
        self.wordchars = ('abcdfeghijklmnopqrstuvwxyz'
                          'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_'
                          'ßàáâãäåæçèéêëìíîïðñòóôõöøùúûüýþÿ'
                          'ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖØÙÚÛÜÝÞ'
                          '!@#$%^&*()_+:;?/>.<,~`|\{[}]-=')
        self.whitespace = ' \t\r\n'
        self.escapedquotes = '"'
        self.state = ' '
        self.pushback = deque()
        self.token = ''

        self.singlequote = "'"
        self.doublequote = '"'
        self.multilinequote = '\n;\n'
        
        self.prevchar = ''
        self.nextchar = ''

        # stream position gets incremented by 1 each time instream.read(1) is called
        # since instream has 0-based numeration, the first emitted character
        # gets streamposition = 0, hence initial streamposition = -1
        self.streamposition = -1
        self.streamlength = len(self.instream.getvalue())
        # self.streamlength = len(self.instream)
 
    def get_token(self):
        """Get a token from the input stream (or from stack if it's nonempty).

        :return: current token
        :rtype: str
        """
        if self.pushback:
            token = self.pushback.popleft()
            return token
        # No pushback.  Get a token.
        raw = self.read_token()
        # print("from get_token:", raw)
        return raw

    def read_token(self):
        """Read token based on the parsing rules

        :return: current token
        :rtype: str
        """
        quoted = False

        # next streamposition, i.e. streamposition+1 should not exceed the number of
        # characters inside instream, streamlength-1 ==> stopping criteria for the while loop
        
        while self.streamposition+1 < self.streamlength-1:
            self.prevchar = self.nextchar

            self.nextchar = self.instream.read(1)
            self.streamposition += 1

            nextnextchar = self.instream.read(1)       # look up 1 char ahead
            self.instream.seek(self.streamposition+1)  # return to current stream position

            # print(repr(self.prevchar), repr(self.nextchar), repr(nextnextchar))
            if self.prevchar+self.nextchar+nextnextchar == '\n;\n': #self.multilinequote:
                self.state = self.multilinequote

            if self.state is None:
                self.token = ''        # past end of file
                break

            elif self.state == ' ':
                if not self.nextchar:
                    self.state = None  # end of file
                    break
                elif self.nextchar in self.whitespace:
                    if self.token:
                        break   # emit current token
                    else:
                        continue
                elif self.nextchar in self.commenters:
                    line = self.instream.readline()
                    self.streamposition += len(line)
                elif self.nextchar in self.wordchars:
                    self.token = self.nextchar
                    self.state = 'a'
                elif self.nextchar in self.singlequote:
                    self.token = self.nextchar
                    self.state = self.nextchar
                elif self.nextchar in self.doublequote:
                    self.token = self.nextchar
                    self.state = self.nextchar
                # elif self.nextchar in self.multilinequote:
                #     self.token = self.nextchar
                #     self.state = self.nextchar
                else:
                    self.token = self.nextchar
                    if self.token:
                        break   # emit current token
                    else:
                        continue

            # Process multiline-quoted text
            elif self.state == self.multilinequote:
                quoted = True
                if not self.nextchar:      # end of file
                    raise EOFError("No closing quotation")

                line = self.instream.readline()
                self.streamposition += len(line) # skip first line == ';\n'

                line = self.instream.readline()
                self.streamposition += len(line)

                while True:
                    if line.startswith(';'): # end of multiline string
                        # self.token = " ".join(self.token.split()) # clean up multiline string - remove extra ' ' and '\n' characters
                        # self.token = ';' + self.token + ';'
                        self.prevchar = '\n'             # if line starts with ';' previous must be '\n'
                        self.streamposition -= len(line) # get back to the beginning of line
                        self.nextchar = self.instream.read(1) # emit self.nextchar after ';'
                        self.streamposition += 1
                        self.state = ' '
                        break   # emit current token
                    else:
                        self.token = self.token + line   # continue concatenating lines to token
                        line = self.instream.readline()
                        self.streamposition += len(line)

            # process token staring with single quote '
            elif self.state in self.singlequote:
                quoted = True
                if not self.nextchar:      # end of file
                    raise EOFError("No closing quotation")

                if self.nextchar == self.state:
                    if nextnextchar not in self.whitespace:
                        self.token = self.token + self.nextchar
                        self.state = self.singlequote
                    elif nextnextchar in self.whitespace:
                        self.token = self.token + self.nextchar
                        self.state = ' '
                        break   # emit current token
                else:
                    self.token = self.token + self.nextchar

            # process token staring with double quote "
            elif self.state in self.doublequote:
                quoted = True
                if not self.nextchar:      # end of file
                    raise EOFError("No closing quotation")
                if self.nextchar == self.state:
                    if nextnextchar not in self.whitespace:
                        self.token = self.token + self.nextchar
                        self.state = self.doublequote
                    elif nextnextchar in self.whitespace:
                        self.token = self.token + self.nextchar
                        self.state = ' '
                        break
                else:
                    self.token = self.token + self.nextchar

            elif self.state == 'a':
                if not self.nextchar:
                    self.state = None   # end of file
                    break
                elif self.nextchar in self.whitespace:
                    self.state = ' '
                    if self.token or quoted:
                        break   # emit current token
                    else:
                        continue
                else:
                    self.token = self.token + self.nextchar

        result = self.token
        self.token = ''
        return result

    def error_leader(self, infile=None, lineno=None):
        """Emit a C-compiler-like, Emacs-friendly error-message leader."""
        if infile is None:
            infile = self.infile
        # if lineno is None:
        #     lineno = self.lineno
        # return "\"%s\", line %d: " % (infile, lineno)
        return ''

    def __iter__(self):
        return self

    def __next__(self):
        token = self.get_token()
        if token == self.eof:
            raise StopIteration
        return token

if __name__ == '__main__':
    if len(sys.argv) == 1:
        lexer = bmrblex()
    else:
        file = sys.argv[1]
        lexer = bmrblex(open(file), file)
    while 1:
        tt = lexer.get_token()
        if tt:
            print("Token: " + repr(tt))
        else:
            break


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
nmrstarlib.csviewer
~~~~~~~~~~~~~~~~~~~

This module provides the :func:`~nmrstarlib.csviewer.csviewer` function - Chemical Shifts Viewer
that visualizes chemical shifts values.
"""

from graphviz import Source
from . import nmrstarlib

class CSViewer(object):
    """Chemical Shifts Viewer: uses:meth:`nmrstarlib.nmrstarlib.StarFile.chem_shifts_by_residue`
    method chemical shifts organized by residue and visualizes chemical shifts values using the
    Graphviz (http://www.graphviz.org/) DOT Languge description.
    """

    dot_template = '''digraph {{
        compound=true
        fontname="Inconsolata, Consolas"
        fontsize=10
        margin="0,0"
        ranksep=0.2
        rankdir=LR
        penwidth=0.5

        node [fontname="Inconsolata, Consolas", fontsize=10, penwidth=0.5]
        edge [fontname="Inconsolata, Consolas", fontsize=10]

        subgraph cluster1 {{
            margin="10,10"
            labeljust="left"
            node [shape=Mrecord, style=filled, colorscheme=spectral11]

            {{
    {}
            }}

            edge [arrowhead=none]
    {}
        }}
    }}
    '''

    def __init__(self, from_path, aminoacids=None, atoms=None, filename=None, csview_format='svg', nmrstarversion=3):
        """CSViewer initializer.

        :param str from_path: Path to single NMR-STAR file or BMRB id.
        :param list aminoacids: List of atom types, e.g. 'ALA', 'GLY', 'SER', etc. Leave as `None` to include everything.
        :param list atoms: List of atom types, e.g. 'CA', 'CB', 'HA', etc. Leave as `None` to include everything.
        :param str filename: Output filename chemical shifts graph to be saved.
        :param str format: `svg`, `png`, `pdf`. See http://www.graphviz.org/doc/info/output.html for all available formats.
        :return: None
        :rtype: None
        """
        self.from_path = from_path
        self.aminoacids = aminoacids
        self.atoms = atoms
        self.filename = filename
        self.csview_format = csview_format
        self.nmrstarversion = nmrstarversion

    def csview(self, view=False):
        """View chemical shift values organized by amino acid residue.

        :param bool view: Open in default image viewer or save file in current working directory quietly.
        :return: None
        :rtype: None
        """
        for starfile in nmrstarlib.read_files([self.from_path]):
            chains = starfile.chem_shifts_by_residue(self.aminoacids, self.atoms, self.nmrstarversion)
            for idx, chemshifts_dict in enumerate(chains):
                nodes = []
                edges = []

                for aminoacid in chemshifts_dict:
                    aaname = '{}_{}'.format(aminoacid[1], aminoacid[0])
                    label = '"{{{}|{}}}"'.format(aminoacid[0], aminoacid[1])
                    color = 8
                    aanode_entry = "            {} [label={}, fillcolor={}]".format(aaname, label, color)
                    nodes.append(aanode_entry)
                    currnodename = aaname

                    for atomtype in chemshifts_dict[aminoacid]:
                        atname = "{}_{}".format(aaname, atomtype)
                        label = '"{{{}|{}}}"'.format(atomtype, chemshifts_dict[aminoacid][atomtype])
                        if atomtype.startswith("H"):
                            color = 4
                        elif atomtype.startswith("C"):
                            color = 6
                        elif atomtype.startswith("N"):
                            color = 10
                        else:
                            color = 8
                        atnode_entry = "{} [label={}, fillcolor={}]".format(atname, label, color)
                        nextnodename = atname
                        nodes.append(atnode_entry)
                        edges.append("{} -> {}".format(currnodename, nextnodename))
                        currnodename = nextnodename

                if self.filename is None:
                    filename = "{}_{}".format(starfile.bmrbid, idx)
                else:
                    filename = "{}_{}".format(self.filename, idx)

                src = Source(self.dot_template.format("\n".join(nodes), "\n".join(edges)), format=self.csview_format)
                src.render(filename=filename, view=view)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
nmrstarlib command-line interface

Usage:
    nmrstarlib -h | --help
    nmrstarlib --version
    nmrstarlib convert (<from_path> <to_path>) [--from_format=<format>] [--to_format=<format>] [--bmrb_url=<url>] [--nmrstarversion=<version>] [--verbose]
    nmrstarlib csview <starfile_path> [--aminoacids=<aa>] [--atoms=<at>] [--csview_outfile=<path>] [--csview_format=<format>] [--nmrstarversion=<version>] [--verbose]

Options:
    -h, --help                   Show this screen.
    --version                    Show version.
    --verbose                    Print what files are processing.
    --from_format=<format>       Input file format, available formats: nmrstar, json [default: nmrstar]
    --to_format=<format>         Output file format, available formats: nmrstar, json [default: json]
    --nmrstarversion=<version>   Version of NMR-STAR format to use, available: 3, 2 [default: 3]
    --bmrb_url=<url>             URL to BMRB REST interface [default: http://rest.bmrb.wisc.edu/bmrb/NMR-STAR3/]
    --aminoacids=<aa>            Comma-separated amino acid three-letter codes
    --atoms=<at>                 Comma-separated BMRB atom codes
    --csview_outfile=<path>      Where to save chemical shifts table
    --csview_format=<format>     Format to which save chamical shift table [default: svg]
"""

import docopt

from . import nmrstarlib
from .converter import Converter
from .csviewer import CSViewer


def main(cmdargs):

    if cmdargs["convert"]:
        nmrstarlib.BMRB_REST = cmdargs["--bmrb_url"]
        nmrstarlib.VERBOSE = cmdargs["--verbose"]
        nmrstarlib.NMRSTAR_VERSION = cmdargs["--nmrstarversion"]

        nmrstarconverter = Converter(from_path=cmdargs["<from_path>"], to_path=cmdargs["<to_path>"],
                                     from_format=cmdargs['--from_format'], to_format=cmdargs["--to_format"])
        nmrstarconverter.convert()

    elif cmdargs["csview"]:
        aminoacids = cmdargs["--aminoacids"].split(",") if cmdargs["--aminoacids"] else []
        atoms = cmdargs["--atoms"].split(",") if cmdargs["--atoms"] else []

        csviewer = CSViewer(from_path=cmdargs["<starfile_path>"], aminoacids=aminoacids, atoms=atoms,
                            filename=cmdargs["--csview_outfile"], csview_format=cmdargs["--csview_format"])
        csviewer.csview(view=True)


args = docopt.docopt(__doc__)
main(args)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
nmrstarlib.bmrblex
~~~~~~~~~~~~~~~~~~

A lexical analyzer for BMRB NMR-STAR format syntax.

This module is based on python ``shlex`` module modified to address specifics of BMRB NMR-STAR format.
The ``shlex`` class makes it easy to write lexical analyzers for simple syntaxes resembling
that of the Unix shell. Documentation: https://docs.python.org/3/library/shlex.html


Simplified description of parsing rules:
----------------------------------------
   * Each word or number separated by whitespace characters is a separate BMRB token.
   * Each single quoted (') string is a separate BMRB token, it should start with a single quote (')
     and end with a single quote *always* followed by whitespace character(s).
   * Each double quoted (") string is a separate BMRB token, it should start with a double quote (")
     and end with a double quote *always* followed by whitespace character(s).
   * Single quoted and double quoted strings have to be processed separately.
   * Single quoted and double quoted strings are processed one character at a time.
   * Multiline strings start with a semicolon *always* followed by new line character and
     ending with a semicolon *always* followed by whitespace character(s).
   * Multiline strings are processed one line at a time.

.. note::
   * For a full description of NMR-STAR file format, see official documentation:
     http://www.bmrb.wisc.edu/dictionary/
   * For a concise description of the NMR-STAR file format grammar see:
     https://github.com/mattfenwick/NMRPyStar#nmr-star-grammar
"""

import sys
from collections import deque
from io import StringIO

__all__ = ["bmrblex"]


class bmrblex:
    """A lexical analyzer for the BMRB NMR-STAR format syntax."""

    multilinequote = 'ಠ'

    def __init__(self, instream=None):
        """Initialize stream string and attributes.

        :param instream: Input string to be converted into stream object for tokens processing.
        :type instream: str or bytes
        """
        if isinstance(instream, str):
            text = self._transform_text(instream)
        elif isinstance(instream, bytes):
            text = self._transform_text(instream.decode("utf-8"))
        else:
            raise TypeError("Expecting <class 'str'> or <class 'bytes'>, but {} was passed".format(type(instream)))

        instream = StringIO(text)

        if instream is not None:
            self.instream = instream
        else:
            self.instream = sys.stdin

        self.eof = ''
        self.commenters = '#'
        self.wordchars = ('abcdfeghijklmnopqrstuvwxyz'
                          'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_'
                          'ßàáâãäåæçèéêëìíîïðñòóôõöøùúûüýþÿ'
                          'ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖØÙÚÛÜÝÞ'
                          '!@#$%^&*()_+:;?/>.<,~`|\{[}]-=')
        self.whitespace = ' \t\r\n'
        self.escapedquotes = '"'
        self.state = ' '
        self.pushback = deque()
        self.token = ''
        self.singlequote = "'"
        self.doublequote = '"'

        # stream position gets incremented by 1 each time instream.read(1) is called
        # since instream has 0-based numeration, the first emitted character
        # gets streamposition = 0, hence initial streamposition = -1
        self.streamposition = -1
        self.streamlength = len(text)

    def _transform_text(self, text):
        """Replace all lines that start with ';' with a special character
        to simplify multiline string processing with bmrblex.

        :param str text: Original text.
        :return: Transformed text where lines starting with ';' replaced by a special character.
        :rtype: str
        """
        lines = text.split('\n')
        newtext = ''
        for line in lines:
            if line and line.startswith(';'):
                line = self.multilinequote + line[1:]
            newtext += line + '\n'
        return newtext
 
    def get_token(self):
        """Get a token from the input stream (or from stack if it's nonempty).

        :return: Current token.
        :rtype: str
        """
        if self.pushback:
            token = self.pushback.popleft()
            return token
        # No pushback.  Get a token.
        raw = self._read_token()
        return raw

    def _read_token(self):
        """Read token based on the parsing rules.

        :return: Current token.
        :rtype: str
        """
        quoted = False

        # next streamposition, i.e. streamposition+1 should not exceed the number of
        # characters inside instream, streamlength-1 ==> stopping criteria for the while loop
        while self.streamposition+1 < self.streamlength-1:
            nextchar = self.instream.read(1)
            self.streamposition += 1

            nextnextchar = self.instream.read(1)       # look up 1 char ahead
            self.instream.seek(self.streamposition+1)  # return to current stream position

            if self.state is None:
                self.token = ""        # past end of file
                break

            elif self.state == " ":
                if not nextchar:
                    self.state = None  # end of file
                    break
                elif nextchar in self.whitespace:
                    if self.token:
                        break   # emit current token
                    else:
                        continue
                elif nextchar in self.commenters:
                    line = self.instream.readline()
                    self.streamposition += len(line)
                elif nextchar in self.wordchars:
                    self.token = nextchar
                    self.state = 'a'
                elif nextchar in self.singlequote:
                    self.token = nextchar
                    self.state = nextchar
                elif nextchar in self.doublequote:
                    self.token = nextchar
                    self.state = nextchar
                elif nextchar in self.multilinequote:
                    # self.token = nextchar
                    self.state = nextchar
                else:
                    self.token = nextchar
                    if self.token:
                        break   # emit current token
                    else:
                        continue

            # Process multiline-quoted text
            elif self.state == self.multilinequote:
                quoted = True
                if not nextchar:      # end of file
                    raise EOFError("No closing quotation")

                line = self.instream.readline()
                self.streamposition += len(line)

                while True:
                    if line.startswith('ಠ'):
                        self.streamposition -= len(line)  # get back to the beginning of line
                        nextchar = self.instream.read(1)  # emit nextchar after 'ಠ'
                        self.streamposition += 1
                        self.token = '\n;\n' + self.token + ';'
                        self.state = ' '
                        break   # emit current token
                    else:
                        self.token = self.token + line
                        line = self.instream.readline()
                        self.streamposition += len(line)

            # process token staring with single quote '
            elif self.state in self.singlequote:
                quoted = True
                if not nextchar:      # end of file
                    raise EOFError("No closing quotation")

                if nextchar == self.state:
                    if nextnextchar not in self.whitespace:
                        self.token = self.token + nextchar
                        self.state = self.singlequote
                    elif nextnextchar in self.whitespace:
                        self.token = self.token + nextchar
                        self.state = ' '
                        break   # emit current token
                else:
                    self.token = self.token + nextchar

            # process token staring with double quote "
            elif self.state in self.doublequote:
                quoted = True
                if not nextchar:      # end of file
                    raise EOFError("No closing quotation")
                if nextchar == self.state:
                    if nextnextchar not in self.whitespace:
                        self.token = self.token + nextchar
                        self.state = self.doublequote
                    elif nextnextchar in self.whitespace:
                        self.token = self.token + nextchar
                        self.state = ' '
                        break
                else:
                    self.token = self.token + nextchar

            # process regular token (unquoted text)
            elif self.state == "a":
                if not nextchar:
                    self.state = None   # end of file
                    break
                elif nextchar in self.whitespace:
                    self.state = " "
                    if self.token or quoted:
                        break   # emit current token
                    else:
                        continue
                else:
                    self.token = self.token + nextchar

        result = self.token
        self.token = ''
        return result

    def __iter__(self):
        return self

    def __next__(self):
        token = self.get_token()
        if token == self.eof:
            raise StopIteration
        return token



#!/usr/bin/env python3

import sys
import os
import time
from . import nmrstarlib

def reading_timetest(dirpath):
    starfiles_dir_path = os.path.abspath(dirpath)
    print("Reading from {}".format(starfiles_dir_path))

    filenames = os.listdir(starfiles_dir_path)
    print("Found {} files".format(len(filenames)))

    ta = time.time()
    print("START:\r\t\t", ta)

    for i, f in enumerate(filenames):
        fta = time.time()
        try:
            sfg = nmrstarlib.read_files([os.path.join(starfiles_dir_path, f)])
            sf = next(sfg)
            ftb = time.time()
            print("OKAY\t" + f + "\t", ftb - fta, "\t", i)
        except KeyboardInterrupt:
            exit()
        except:
            ftb = time.time()
            print("FAIL\t" + f + "\t", ftb - fta, "\t", i)

    tb = time.time()
    print("END:\r\t\t", tb)
    print("Files read in {} seconds".format(tb - ta))

if __name__ == "__main__":
    # python3 -m nmrstarlib.timetest ../AutomatedAnalysisMASSSNMR/bmrbscraper/NMR-STAR3

    script = sys.argv.pop(0)
    starfiles_dir_path = sys.argv.pop(0)

    reading_timetest(starfiles_dir_path)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import cProfile
from . import nmrstarlib


if __name__ == "__main__":
    # python3 -m nmrstarlib.profile > profile.txt 15573

    source = sys.argv[1:]
    starfile_generator = nmrstarlib.read_files(source)
    cProfile.run("next(starfile_generator)")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
nmrstarlib.converter
~~~~~~~~~~~~~~~~~~~~

This module provides functionality for converting between the BMRB
NMR-STAR format and its equivalent JSONized NMR-STAR format.

The following conversions are possible:

Local files:
   * One-to-one file conversions:
      * textfile - to - textfile
      * textfile - to - textfile.gz
      * textfile - to - textfile.bz2
      * textfile.gz - to - textfile
      * textfile.gz - to - textfile.gz
      * textfile.gz - to - textfile.bz2
      * textfile.bz2 - to - textfile
      * textfile.bz2 - to - textfile.gz
      * textfile.bz2 - to - textfile.bz2
      * textfile / textfile.gz / textfile.bz2 - to - textfile.zip / textfile.tar / textfile.tar.gz / textfile.tar.bz2 (TypeError: One-to-many conversion)
   * Many-to-many files conversions:
      * Directories:
         * directory - to - directory
         * directory - to - directory.zip
         * directory - to - directory.tar
         * directory - to - directory.tar.bz2
         * directory - to - directory.tar.gz
         * directory - to - directory.gz / directory.bz2 (TypeError: Many-to-one conversion)
      * Zipfiles:
         * zipfile.zip - to - directory
         * zipfile.zip - to - zipfile.zip
         * zipfile.zip - to - tarfile.tar
         * zipfile.zip - to - tarfile.tar.gz
         * zipfile.zip - to - tarfile.tar.bz2
         * zipfile.zip - to - directory.gz / directory.bz2 (TypeError: Many-to-one conversion)
      * Tarfiles:
         * tarfile.tar - to - directory
         * tarfile.tar - to - zipfile.zip
         * tarfile.tar - to - tarfile.tar
         * tarfile.tar - to - tarfile.tar.gz
         * tarfile.tar - to - tarfile.tar.bz2
         * tarfile.tar - to - directory.gz / directory.bz2 (TypeError: Many-to-one conversion)
         * tarfile.tar.gz - to - directory
         * tarfile.tar.gz - to - zipfile.zip
         * tarfile.tar.gz - to - tarfile.tar
         * tarfile.tar.gz - to - tarfile.tar.gz
         * tarfile.tar.gz - to - tarfile.tar.bz2
         * tarfile.tar.gz - to - directory.gz / directory.bz2 (TypeError: Many-to-one conversion)
         * tarfile.tar.bz2 - to - directory
         * tarfile.tar.bz2 - to - zipfile.zip
         * tarfile.tar.bz2 - to - tarfile.tar
         * tarfile.tar.bz2 - to - tarfile.tar.gz
         * tarfile.tar.bz2 - to - tarfile.tar.bz2
         * tarfile.tar.bz2 - to - directory.gz / directory.bz2 (TypeError: Many-to-one conversion)
URL files:
   * One-to-one file conversions:
      * bmrbid - to - textfile
      * bmrbid - to - textfile.gz
      * bmrbid - to - textfile.bz2
      * bmrbid - to - textfile.zip / textfile.tar / textfile.tar.gz / textfile.tar.bz2 (TypeError: One-to-many conversion)
      * textfileurl - to - textfile
      * textfileurl - to - textfile.gz
      * textfileurl - to - textfile.bz2
      * textfileurl.gz - to - textfile
      * textfileurl.gz - to - textfile.gz
      * textfileurl.gz - to - textfile.bz2
      * textfileurl.bz2 - to - textfile
      * textfileurl.bz2 - to - textfile.gz
      * textfileurl.bz2 - to - textfile.bz2
      * textfileurl / textfileurl.gz / textfileurl.bz2 - to - textfile.zip / textfile.tar / textfile.tar.gz / textfile.tar.bz2 (TypeError: One-to-many conversion)
   * Many-to-many files conversions:
      * Zipfiles:
         * zipfileurl.zip - to - directory
         * zipfileurl.zip - to - zipfile.zip
         * zipfileurl.zip - to - tarfile.tar
         * zipfileurl.zip - to - tarfile.tar.gz
         * zipfileurl.zip - to - tarfile.tar.bz2
         * zipfileurl.zip - to - directory.gz / directory.bz2 (TypeError: Many-to-one conversion)
      * Tarfiles:
         * tarfileurl.tar - to - directory
         * tarfileurl.tar - to - zipfile.zip
         * tarfileurl.tar - to - tarfile.tar
         * tarfileurl.tar - to - tarfile.tar.gz
         * tarfileurl.tar - to - tarfile.tar.bz2
         * tarfileurl.tar - to - directory.gz / directory.bz2 (TypeError: Many-to-one conversion)
         * tarfileurl.tar.gz - to - directory
         * tarfileurl.tar.gz - to - zipfile.zip
         * tarfileurl.tar.gz - to - tarfile.tar
         * tarfileurl.tar.gz - to - tarfile.tar.gz
         * tarfileurl.tar.gz - to - tarfile.tar.bz2
         * tarfileurl.tar.gz - to - directory.gz / directory.bz2 (TypeError: Many-to-one conversion)
         * tarfileurl.tar.bz2 - to - directory
         * tarfileurl.tar.bz2 - to - zipfile.zip
         * tarfileurl.tar.bz2 - to - tarfile.tar
         * tarfileurl.tar.bz2 - to - tarfile.tar.gz
         * tarfileurl.tar.bz2 - to - tarfile.tar.bz2
         * tarfileurl.tar.bz2 - to - directory.gz / directory.bz2 (TypeError: Many-to-one conversion)
"""

import os
import io
import zipfile
import tarfile
import bz2
import gzip

from . import nmrstarlib


class Converter(object):
    """Converter class to convert BMRB NMR-STAR files from NMR-STAR to JSON or from JSON to NMR-STAR format."""

    nmrstar_extension = {"json":    ".json",
                         "nmrstar": ".str"}

    def __init__(self, from_path, to_path, from_format="nmrstar", to_format="json"):
        """Converter initializer.

        :param str from_path: Path to input file(s).
        :param str to_path: Path to output file(s).
        :param str from_format: Input format: `nmrstar` or `json`.
        :param str to_format: Output format: `nmrstar` or `json`.
        :return: None
        :rtype: None
        """
        self.from_path = os.path.normpath(from_path)
        self.to_path = os.path.normpath(to_path)
        self.from_format = from_format
        self.to_format = to_format

        self.from_path_compression = nmrstarlib.GenericFilePath.is_compressed(from_path)
        self.to_path_compression = nmrstarlib.GenericFilePath.is_compressed(to_path)

    def convert(self):
        """Convert file(s) from NMR-STAR format to JSON format or from JSON format to NMR-STAR format.

        :return: None
        :rtype: None
        """
        if not os.path.exists(os.path.dirname(self.to_path)):
            dirname = os.path.dirname(self.to_path)
            if dirname:
                os.makedirs(dirname)

        if os.path.isdir(self.from_path):
            self._many_to_many()
        elif os.path.isfile(self.from_path) or nmrstarlib.GenericFilePath.is_url(self.from_path):
            if self.from_path_compression in ("zip", "tar", "tar.gz", "tar.bz2"):
                self._many_to_many()
            elif self.from_path_compression in ("gz", "bz2"):
                self._one_to_one()
            elif not self.from_path_compression:
                self._one_to_one()
        elif self.from_path.isdigit():
            self._one_to_one()
        else:
            raise TypeError('Unknown input file format: "{}"'.format(self.from_path))

    def _many_to_many(self):
        """Perform many-to-many files conversion.

        :return: None
        :rtype: None
        """
        if not self.to_path_compression:
            self._to_dir()
        elif self.to_path_compression == "zip":
            self._to_zipfile()
        elif self.to_path_compression in ("tar", "tar.gz", "tar.bz2"):
            self._to_tarfile()
        elif self.to_path_compression in ("gz", "bz2"):
            raise TypeError('Many-to-one conversion, cannot convert "{}" into "{}"'.format(self.from_path,
                                                                                           self.to_path))
        else:
            raise TypeError('Unknown output file format: "{}"'.format(self.to_path))

    def _one_to_one(self):
        """Perform one-to-one file conversion.

        :return: None
        :rtype: None
        """
        if not self.to_path_compression:
            self._to_textfile()
        elif self.to_path_compression == "gz":
            self._to_gzipfile()
        elif self.to_path_compression == "bz2":
            self._to_bz2file()
        elif self.to_path_compression in ("tar", "tar.gz", "tar.bz2", "zip"):
            raise TypeError('One-to-many conversion, cannot convert "{}" into "{}"'.format(self.from_path,
                                                                                           self.to_path))
        else:
            raise TypeError('Unknown format: "{}"'.format(self.to_path))

    def _to_dir(self):
        """Convert files to directory.

        :return: None
        :rtype: None
        """
        for starfile in nmrstarlib.read_files([self.from_path]):
            outpath = self._outputpath(starfile.source)

            if not os.path.exists(os.path.dirname(outpath)):
                os.makedirs(os.path.dirname(outpath))

            with open(outpath, mode="w") as outfile:
                starfile.write(outfile, self.to_format)

    def _to_zipfile(self):
        """Convert files to zip archive.

        :return: None
        :rtype: None
        """
        with zipfile.ZipFile(self.to_path, mode="w", compression=zipfile.ZIP_DEFLATED) as outfile:
            for starfile in nmrstarlib.read_files([self.from_path]):
                outpath = self._outputpath(starfile.source, archive=True)
                outfile.writestr(outpath, starfile.writestr(self.to_format))

    def _to_tarfile(self):
        """Convert files to tar archive.

        :return: None
        :rtype: None
        """
        if self.to_path_compression == "tar":
            tar_mode = "w"
        elif self.to_path_compression == "tar.gz":
            tar_mode = "w:gz"
        elif self.to_path_compression == 'tar.bz2':
            tar_mode = "w:bz2"
        else:
            tar_mode = "w"

        with tarfile.open(self.to_path, mode=tar_mode) as outfile:
            for starfile in nmrstarlib.read_files([self.from_path]):
                outpath = self._outputpath(starfile.source, archive=True)
                info = tarfile.TarInfo(outpath)
                data = starfile.writestr(self.to_format).encode()
                info.size = len(data)
                outfile.addfile(tarinfo=info, fileobj=io.BytesIO(data))

    def _to_bz2file(self):
        """Convert file to bz2-compressed file.

        :return: None
        :rtype: None
        """
        with bz2.BZ2File(self.to_path, mode="wb") as outfile:
            for starfile in nmrstarlib.read_files([self.from_path]):
                outfile.write(starfile.writestr(self.to_format).encode())

    def _to_gzipfile(self):
        """Convert file to gzip-compressed file.

        :return: None
        :rtype: None
        """
        with gzip.GzipFile(self.to_path, mode="wb") as outfile:
            for starfile in nmrstarlib.read_files([self.from_path]):
                outfile.write(starfile.writestr(self.to_format).encode())

    def _to_textfile(self):
        """Convert file to regular text file.

        :return: None
        :rtype: None
        """
        to_path = self.to_path if self.to_path.endswith(self.nmrstar_extension[self.to_format]) \
            else self.to_path + self.nmrstar_extension[self.to_format]
        with open(to_path, mode="w") as outfile:
            for starfile in nmrstarlib.read_files([self.from_path]):
                outfile.write(starfile.writestr(self.to_format))

    def _outputpath(self, inputpath, archive=False):
        """Construct an output path string from an input path string.

        :param str inputpath: Input path string.
        :return: Output path string.
        :rtype: str
        """
        indirpath, fname = os.path.split(os.path.abspath(os.path.normpath(inputpath)))

        commonprefix = os.path.commonprefix([os.path.abspath(self.from_path),
                                             os.path.abspath(indirpath)])

        commonparts = commonprefix.split(os.sep)
        inparts = indirpath.split(os.sep)
        outparts = inparts[len(commonparts):]

        if archive:
            outdirpath = os.path.join(*outparts)
        else:
            outdirpath = os.path.join(self.to_path, *outparts)

        return os.path.join(outdirpath, fname + self.nmrstar_extension[self.to_format])



# setup.py

from setuptools import setup, find_packages

setup(
    name='nmrstarlib',
    version='1.0.4',
    author='Andrey Smelter',
    author_email='andrey.smelter@gmail.com',
    description='Python library for parsing data from NMR-STAR format files',
    keywords='BMRB NMR-STAR parsing nmrstarlib',
    license='MIT',
    url='https://github.com/MoseleyBioinformaticsLab/nmrstarlib',
    packages=find_packages(),
    package_data={'nmrstarlib':['conf/*.json']},
    platforms='any',
    long_description=open('README.rst').read(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)


#!/usr/bin/env python2

from __future__ import print_function
from __future__ import division

import sys
import os
import time

sys.path.append('../extras/pystarlib/pystarlib/src')
from STAR import File


d = os.path.abspath(sys.argv[1])
print("Reading from %s.  "%(d),end='')


filenames = [os.path.join(d,f) for f in os.listdir(d)]
print("Found %d files."%(len(filenames)))

ta = time.time()
print("START:\r\t\t", ta)

for i, f in enumerate(filenames):
    fta = time.time()
    try:
        sf = File.File(filename=f)
        sf.read()
        ftb = time.time()
        print("OKAY\t" + os.path.basename(f) + "\t", ftb - fta, "\t", i)
    except KeyboardInterrupt:
        exit()
    except:
        ftb = time.time()
        print("FAIL\t" + os.path.basename(f) + "\t", ftb - fta, "\t", i)

tb = time.time()
print("END:\r\t\t", tb)
print("Files read in %s seconds"%(tb-ta))
